@@ -22,10 +22,6 @@ from resize_right import resize
 
 from imagen_pytorch.t5 import t5_encode_text, get_encoded_dim, DEFAULT_T5_NAME
 
-# constants
-
-NAT = 1. / math.log(2.)
-
 # helper functions
 
 def exists(val):
@@ -134,36 +130,6 @@ def extract(a, t, x_shape):
     out = a.gather(-1, t)
     return out.reshape(b, *((1,) * (len(x_shape) - 1)))
 
-def meanflat(x):
-    return x.mean(dim = tuple(range(1, len(x.shape))))
-
-def normal_kl(mean1, logvar1, mean2, logvar2):
-    return 0.5 * (-1.0 + logvar2 - logvar1 + torch.exp(logvar1 - logvar2) + ((mean1 - mean2) ** 2) * torch.exp(-logvar2))
-
-def approx_standard_normal_cdf(x):
-    return 0.5 * (1.0 + torch.tanh(((2.0 / math.pi) ** 0.5) * (x + 0.044715 * (x ** 3))))
-
-def discretized_gaussian_log_likelihood(x, *, means, log_scales, thres = 0.999):
-    assert x.shape == means.shape == log_scales.shape
-
-    centered_x = x - means
-    inv_stdv = torch.exp(-log_scales)
-    plus_in = inv_stdv * (centered_x + 1. / 255.)
-    cdf_plus = approx_standard_normal_cdf(plus_in)
-    min_in = inv_stdv * (centered_x - 1. / 255.)
-    cdf_min = approx_standard_normal_cdf(min_in)
-    log_cdf_plus = log(cdf_plus)
-    log_one_minus_cdf_min = log(1. - cdf_min)
-    cdf_delta = cdf_plus - cdf_min
-
-    log_probs = torch.where(x < -thres,
-        log_cdf_plus,
-        torch.where(x > thres,
-            log_one_minus_cdf_min,
-            log(cdf_delta)))
-
-    return log_probs
-
 def cosine_beta_schedule(timesteps, s = 0.008, thres = 0.999):
     """
     cosine schedule
@@ -182,7 +148,6 @@ def linear_beta_schedule(timesteps):
     beta_end = scale * 0.02
     return torch.linspace(beta_start, beta_end, timesteps, dtype = torch.float64)
 
-
 class GaussianDiffusion(nn.Module):
     def __init__(self, *, beta_schedule, timesteps):
         super().__init__()
@@ -248,17 +213,6 @@ class GaussianDiffusion(nn.Module):
 
         return time_transitions
 
-    def get_learned_posterior_log_variance(self, var_interp_frac_unnormalized, x_t, t):
-        # if learned variance, posterior variance and posterior log variance are predicted by the network
-        # by an interpolation of the max and min log beta values
-        # eq 15 - https://arxiv.org/abs/2102.09672
-        min_log = extract(self.posterior_log_variance_clipped, t, x_t.shape)
-        max_log = extract(torch.log(self.betas), t, x_t.shape)
-        var_interp_frac = unnormalize_zero_to_one(var_interp_frac_unnormalized)
-
-        posterior_log_variance = var_interp_frac * max_log + (1 - var_interp_frac) * min_log
-        return posterior_log_variance
-
     def q_posterior(self, x_start, x_t, t, **kwargs):
         posterior_mean = (
             extract(self.posterior_mean_coef1, t, x_t.shape) * x_start +
@@ -291,7 +245,7 @@ def beta_linear_log_snr(t):
 
 @torch.jit.script
 def alpha_cosine_log_snr(t, s: float = 0.008):
-    return -log((torch.cos((t + s) / (1 + s) * math.pi * 0.5) ** -2) - 1)
+    return -log((torch.cos((t + s) / (1 + s) * math.pi * 0.5) ** -2) - 1, eps = 1e-5)
 
 def log_snr_to_alpha_sigma(log_snr):
     return torch.sqrt(torch.sigmoid(log_snr)), torch.sqrt(torch.sigmoid(-log_snr))
@@ -324,9 +278,6 @@ class GaussianDiffusionContinuousTimes(nn.Module):
         times = times.unbind(dim = -1)
         return times
 
-    def get_learned_posterior_log_variance(self, var_interp_frac_unnormalized, x_t, t):
-        raise NotImplementedError
-
     def q_posterior(self, x_start, x_t, t, *, t_next = None):
         t_next = default(t_next, lambda: (t - 1. / self.num_timesteps).clamp(min = 0.))
 
@@ -1257,8 +1208,6 @@ class Imagen(nn.Module):
         beta_schedules = 'cosine',
         lowres_sample_noise_level = 0.2,            # in the paper, they present a new trick where they noise the lowres conditioning image, and at sample time, fix it to a certain level (0.1 or 0.3) - the unets are also made to be conditioned on this noise level
         condition_on_text = True,
-        learned_variance = True,
-        vb_loss_weight = 0.001,
         auto_normalize_img = True,                  # whether to take care of normalizing the image from [0, 1] to [-1, 1] and back automatically - you can turn this off if you want to pass in the [-1, 1] ranged image yourself from the dataloader
         continuous_times = True,
         dynamic_thresholding_percentile = 0.9       # unsure what this was based on perusal of paper
@@ -1299,10 +1248,6 @@ class Imagen(nn.Module):
         timesteps = cast_tuple(timesteps, num_unets)
         beta_schedules = cast_tuple(beta_schedules, num_unets)
 
-        if continuous_times:
-            # just turn off learned variance for continuous times gaussian diffusion
-            learned_variance = False
-
         noise_scheduler_klass = GaussianDiffusion if not continuous_times else GaussianDiffusionContinuousTimes
         self.noise_schedulers = nn.ModuleList([])
 
@@ -1310,12 +1255,6 @@ class Imagen(nn.Module):
             noise_scheduler = noise_scheduler_klass(beta_schedule = beta_schedule, timesteps = timestep)
             self.noise_schedulers.append(noise_scheduler)
 
-        # whether to use learned variance, defaults to True for the first unet in the cascade, as in paper
-
-        learned_variance = pad_tuple_to_length(cast_tuple(learned_variance), num_unets, fillvalue = False)
-        self.learned_variance = learned_variance
-        self.vb_loss_weight = vb_loss_weight
-
         # get text encoder
 
         self.text_encoder_name = text_encoder_name
@@ -1325,18 +1264,16 @@ class Imagen(nn.Module):
 
         self.unets = nn.ModuleList([])
 
-        for ind, (one_unet, one_unet_learned_var) in enumerate(zip(unets, learned_variance)):
+        for ind, one_unet in enumerate(unets):
             assert isinstance(one_unet, Unet)
             is_first = ind == 0
 
-            unet_channels_out = self.channels * (1 if not one_unet_learned_var else 2)
-
             one_unet = one_unet.cast_model_parameters(
                 lowres_cond = not is_first,
                 cond_on_text = self.condition_on_text,
                 text_embed_dim = self.text_embed_dim if self.condition_on_text else None,
                 channels = self.channels,
-                channels_out = unet_channels_out,
+                channels_out = self.channels,
                 fourier_embed_time_or_noise = not continuous_times
             )
 
@@ -1404,14 +1341,11 @@ class Imagen(nn.Module):
         for unet, device in zip(self.unets, devices):
             unet.to(device)
 
-    def p_mean_variance(self, unet, x, t, *, noise_scheduler, text_embeds = None, text_mask = None, lowres_cond_img = None, lowres_noise_times = None, clip_denoised = True, learned_variance = False, cond_scale = 1., model_output = None, t_next = None):
+    def p_mean_variance(self, unet, x, t, *, noise_scheduler, text_embeds = None, text_mask = None, lowres_cond_img = None, lowres_noise_times = None, clip_denoised = True, cond_scale = 1., model_output = None, t_next = None):
         assert not (cond_scale != 1. and not self.can_classifier_guidance), 'imagen was not trained with conditional dropout, and thus one cannot use classifier free guidance (cond_scale anything other than 1)'
 
         pred = default(model_output, lambda: unet.forward_with_cond_scale(x, noise_scheduler.get_condition(t), text_embeds = text_embeds, text_mask = text_mask, cond_scale = cond_scale, lowres_cond_img = lowres_cond_img, lowres_noise_times = noise_scheduler.get_condition(lowres_noise_times)))
 
-        if learned_variance:
-            pred, var_interp_frac_unnormalized = pred.chunk(2, dim = 1)
-
         x_recon = noise_scheduler.predict_start_from_noise(x, t = t, noise = pred)
 
         if clip_denoised:
@@ -1430,23 +1364,19 @@ class Imagen(nn.Module):
 
         model_mean, posterior_variance, posterior_log_variance = noise_scheduler.q_posterior(x_start = x_recon, x_t = x, t = t, t_next = t_next)
 
-        if learned_variance:
-            posterior_log_variance = noise_scheduler.get_learned_posterior_log_variance(var_interp_frac_unnormalized, x_t = x, t = t)
-            posterior_variance = posterior_log_variance.exp()
-
         return model_mean, posterior_variance, posterior_log_variance
 
     @torch.no_grad()
-    def p_sample(self, unet, x, t, *, noise_scheduler, t_next = None, text_embeds = None, text_mask = None, cond_scale = 1., lowres_cond_img = None, lowres_noise_times = None,  learned_variance = False, clip_denoised = True):
+    def p_sample(self, unet, x, t, *, noise_scheduler, t_next = None, text_embeds = None, text_mask = None, cond_scale = 1., lowres_cond_img = None, lowres_noise_times = None,  clip_denoised = True):
         b, *_, device = *x.shape, x.device
-        model_mean, _, model_log_variance = self.p_mean_variance(unet, x = x, t = t, t_next = t_next, noise_scheduler = noise_scheduler, text_embeds = text_embeds, text_mask = text_mask, cond_scale = cond_scale, lowres_cond_img = lowres_cond_img, lowres_noise_times = lowres_noise_times, clip_denoised = clip_denoised, learned_variance = learned_variance)
+        model_mean, _, model_log_variance = self.p_mean_variance(unet, x = x, t = t, t_next = t_next, noise_scheduler = noise_scheduler, text_embeds = text_embeds, text_mask = text_mask, cond_scale = cond_scale, lowres_cond_img = lowres_cond_img, lowres_noise_times = lowres_noise_times, clip_denoised = clip_denoised)
         noise = torch.randn_like(x)
         # no noise when t == 0
         nonzero_mask = (1 - (t == 0).float()).reshape(b, *((1,) * (len(x.shape) - 1)))
         return model_mean + nonzero_mask * (0.5 * model_log_variance).exp() * noise
 
     @torch.no_grad()
-    def p_sample_loop(self, unet, shape, *, noise_scheduler, learned_variance = False, clip_denoised = True, lowres_cond_img = None, lowres_noise_times = None, text_embeds = None, text_mask = None, cond_scale = 1):
+    def p_sample_loop(self, unet, shape, *, noise_scheduler, clip_denoised = True, lowres_cond_img = None, lowres_noise_times = None, text_embeds = None, text_mask = None, cond_scale = 1):
         device = self.device
 
         batch = shape[0]
@@ -1467,7 +1397,6 @@ class Imagen(nn.Module):
                 lowres_cond_img = lowres_cond_img,
                 lowres_noise_times = lowres_noise_times,
                 noise_scheduler = noise_scheduler,
-                learned_variance = learned_variance,
                 clip_denoised = clip_denoised
             )
 
@@ -1507,7 +1436,7 @@ class Imagen(nn.Module):
 
         lowres_sample_noise_level = default(lowres_sample_noise_level, self.lowres_sample_noise_level)
 
-        for unet_number, unet, channel, image_size, learned_variance, noise_scheduler in tqdm(zip(range(1, len(self.unets) + 1), self.unets, self.sample_channels, self.image_sizes, self.learned_variance, self.noise_schedulers)):
+        for unet_number, unet, channel, image_size, noise_scheduler in tqdm(zip(range(1, len(self.unets) + 1), self.unets, self.sample_channels, self.image_sizes, self.noise_schedulers)):
 
             context = self.one_unet_in_gpu(unet = unet) if is_cuda else null_context()
 
@@ -1529,7 +1458,6 @@ class Imagen(nn.Module):
                     text_embeds = text_embeds,
                     text_mask = text_masks,
                     cond_scale = cond_scale,
-                    learned_variance = learned_variance,
                     clip_denoised = True,
                     lowres_cond_img = lowres_cond_img,
                     lowres_noise_times = lowres_noise_times,
@@ -1545,7 +1473,7 @@ class Imagen(nn.Module):
         pil_images = list(map(T.ToPILImage(), img.unbind(dim = 0)))
         return pil_images # now you have a bunch of pillow images you can just .save(/where/ever/you/want.png)
 
-    def p_losses(self, unet, x_start, times, *, noise_scheduler, lowres_cond_img = None, lowres_aug_times = None, text_embeds = None, text_mask = None, noise = None, learned_variance = False, clip_denoised = False, times_next = None):
+    def p_losses(self, unet, x_start, times, *, noise_scheduler, lowres_cond_img = None, lowres_aug_times = None, text_embeds = None, text_mask = None, noise = None, clip_denoised = False, times_next = None):
         noise = default(noise, lambda: torch.randn_like(x_start))
 
         # normalize to [-1, 1]
@@ -1577,46 +1505,10 @@ class Imagen(nn.Module):
             cond_drop_prob = self.cond_drop_prob,
         )
 
-        if learned_variance:
-            pred, _ = model_output.chunk(2, dim = 1)
-        else:
-            pred = model_output
+        pred = model_output
 
         loss = self.loss_fn(pred, noise)
-
-        if not learned_variance:
-            # return simple loss if not using learned variance
-            return loss
-
-        # most of the code below is transcribed from
-        # https://github.com/hojonathanho/diffusion/blob/master/diffusion_tf/diffusion_utils_2.py
-        # the Improved DDPM paper then further modified it so that the mean is detached (shown a couple lines before), and weighted to be smaller than the l1 or l2 "simple" loss
-        # it is questionable whether this is really needed, looking at some of the figures in the paper, but may as well stay faithful to their implementation
-
-        # if learning the variance, also include the extra weight kl loss
-
-        true_mean, _, true_log_variance_clipped = noise_scheduler.q_posterior(x_start = x_start, x_t = x_noisy, t = times, t_next = times_next)
-        model_mean, _, model_log_variance = self.p_mean_variance(unet, x = x_noisy, t = times, noise_scheduler = noise_scheduler, clip_denoised = clip_denoised, learned_variance = True, model_output = model_output)
-
-        # kl loss with detached model predicted mean, for stability reasons as in paper
-
-        detached_model_mean = model_mean.detach()
-
-        kl = normal_kl(true_mean, true_log_variance_clipped, detached_model_mean, model_log_variance)
-        kl = meanflat(kl) * NAT
-
-        decoder_nll = -discretized_gaussian_log_likelihood(x_start, means = detached_model_mean, log_scales = 0.5 * model_log_variance)
-        decoder_nll = meanflat(decoder_nll) * NAT
-
-        # at the first timestep return the decoder NLL, otherwise return KL(q(x_{t-1}|x_t,x_0) || p(x_{t-1}|x_t))
-
-        vb_losses = torch.where(times == 0, decoder_nll, kl)
-
-        # weight the vb loss smaller, for stability, as in the paper (recommended 0.001)
-
-        vb_loss = vb_losses.mean() * self.vb_loss_weight
-
-        return loss + vb_loss
+        return loss
 
     def forward(
         self,
@@ -1635,7 +1527,6 @@ class Imagen(nn.Module):
         noise_scheduler     = self.noise_schedulers[unet_index]
         target_image_size   = self.image_sizes[unet_index]
         prev_image_size     = self.image_sizes[unet_index - 1] if unet_index > 0 else None
-        learned_variance    = self.learned_variance[unet_index]
         b, c, h, w, device, = *image.shape, image.device
 
         check_shape(image, 'b c h w', c = self.channels)
@@ -1662,4 +1553,4 @@ class Imagen(nn.Module):
 
         image = resize_image_to(image, target_image_size)
 
-        return self.p_losses(unet, image, times, text_embeds = text_embeds, text_mask = text_masks, noise_scheduler = noise_scheduler, lowres_cond_img = lowres_cond_img, lowres_aug_times = lowres_aug_times, learned_variance = learned_variance)
+        return self.p_losses(unet, image, times, text_embeds = text_embeds, text_mask = text_masks, noise_scheduler = noise_scheduler, lowres_cond_img = lowres_cond_img, lowres_aug_times = lowres_aug_times)
